/*
	$Author: me $ - $Date: 1999/10/17 20:20:04 $
*/
/*----------------------------------------------------------------------------------------
	NAT
------------------------------------------------------------------------------------------*/
// Network Address Translation for the BeOS
// Richard Burgess <rickb@spy.net>
// Our basic strategy:
// For outgoing packets, we replace the source IP address with this machine's address and
// assign a new port for each machine on our local network.
// For incoming packets, we use the port number to determine which machine the packet
// is actually ment for and replace its destination address.
// NOTE: all the data read from the packets is kept in network byte-order unless
// otherwise noted
#include <OS.h>
#include <Autolock.h>

#include <netdb.h>
#include <socket.h>
#include <string.h>

#include "nat_config.h"
#include "debug.h"
#include "ip.h"
#include "udp.h"
#include "tcp.h"
#include "ethernet.h"
#include "ppp.h"
#include "protocols.h"
#include "nat.h"

bigtime_t STATIC_PORT_MAP = -1;

// debug file
FILE *f = 0;

// NULL terminated list of funky protocol nat supports
extern nat_protocol funky_protocols[];

nat_config config;

static bool done = false;

/*----------------------------------------------------------------------------------------
	NATHandler
------------------------------------------------------------------------------------------*/
NATHandler::NATHandler() : BPacketHandler(),
		mLocalDev(NULL), mLocalAddr(0), mLocalMask(0),
		mInetDev(NULL),mInetAddr(0)
{	
	done = false;
	
	for(int i =0; i < NAT_NUM_PROTO; i++)
		mFreePort[i] = config.start_port;
		
	// start up our threads
	mSocketReaper 	= spawn_thread(SocketReaper,"Della is Hot!",B_LOW_PRIORITY, this);
	resume_thread(mSocketReaper);
	
	mPortReaper 	= spawn_thread(PortReaper,"www.alleybaggett.net",B_LOW_PRIORITY, this); 	
	resume_thread(mPortReaper);
	
	mTheGreatMaker	= spawn_thread(TheGreatMaker,"Paul is Dead.",B_LOW_PRIORITY, this);
	resume_thread(mTheGreatMaker);
}

/*-----------------------------------------------------------------------------------
	 ~NATHandler
------------------------------------------------------------------------------------------*/
NATHandler::~NATHandler()
{
	// wait for threads to go bye bye
	done = true;
	status_t val;
	
	resume_thread(mSocketReaper);
	resume_thread(mPortReaper);
	resume_thread(mTheGreatMaker);
	
	wait_for_thread (mSocketReaper, &val);
	wait_for_thread (mPortReaper, &val);
	wait_for_thread (mTheGreatMaker, &val);

	if(f) fclose(f);
}

/*-----------------------------------------------------------------------------------
	 AddDevice
------------------------------------------------------------------------------------------*/
// Add our handler to ethernet and ppp devices
void NATHandler::AddDevice(BNetDevice *dev,  const char *name)
{	
	char str[32];
	net_device_type dev_type = dev->Type();
	
	// get local interface, only handling ethernet interfaces
	if( !mLocalDev && (dev_type == B_ETHER_NET_DEVICE) && 
		find_net_setting(NULL,"nat","LOCAL",str,32) )
	{
		if(!strcmp(name, str))
		{
			mLocalDev = dev;
			find_net_setting(NULL, name, "IPADDRESS",str,32);
			mLocalAddr = inet_addr(str);
		
			find_net_setting(NULL, name, "NETMASK",str,32);
			mLocalMask = inet_addr(str);
			
			register_packet_handler(this, dev, 1); 	// we're number 1,
													// so we get the packets first.
		}
	}
	
	// get internet interface, can be ethernet or ppp
	if( !mInetDev && ( (dev_type == B_ETHER_NET_DEVICE) || (dev_type == B_PPP_NET_DEVICE) ) &&
		find_net_setting(NULL,"nat","INTERNET",str,32) )
	{
		if(!strcmp(name, str))
		{
			mInetDev = dev;
			if(dev_type == B_ETHER_NET_DEVICE)
			{
				find_net_setting(NULL, name, "IPADDRESS",str,32);
				mInetAddr = inet_addr(str);
			}
			register_packet_handler(this, dev, 1); 	// we're number 1,
													// so we get the packets first.
		}
	}
}

/*----------------------------------------------------------------------------------------
	 PacketReceived
------------------------------------------------------------------------------------------*/
// We replace the source IP address of these packets which the
// IP address of our machine so the packets will get routed back to us since our local
// machines will not have assigned IP addresses. In order to save the information about
// which machine the packet is actually from, we use a different port number for
// each machine on our network.

bool NATHandler::PacketReceived(BNetPacket *pkt, BNetDevice *dev)
{
	uint16 protocol = 0;
	int32 offset = -1;
	
	switch(dev->Type())
	{
		case B_ETHER_NET_DEVICE:
		{
			pkt->Read(ETHERNET_PROTOCOL_OFFSET, (char*)&protocol, sizeof(protocol));
			
			// only care about ip protocol
			if(ntohs(protocol) == ETHERNET_PROTOID_IP) offset = ETHERNET_HEADER_LENGTH;
		}	
		break;
		
		case B_PPP_NET_DEVICE:
		{
			pkt->Read(PPP_PROTOCOL_OFFSET, (char*)&protocol, sizeof(protocol));
			switch(ntohs(protocol))
			{
				case PPP_IP_PROTOCOL: // ip protocol
					offset = PPP_HEADER_LENGTH;
				break;
				
				case PPP_IPCP_PROTOCOL: // see if ppp ip is in packet
					FindPPPAddress(pkt);	
				break;
			}
		}
		break;
	}
	
	// do our naughty, sinful business
	if(offset > 0)
	{
		pkt->SetBase(offset);
		
		if(dev == mLocalDev)
		{	// need to do our magic on outgoing packets
			OutTranslateIpPacket(pkt);
		}else if (dev == mInetDev)
		{ 	// need to do our magic on incoming packets
			InTranslateIpPacket(pkt);
		}
		
		pkt->SetBase(-offset);
	}
	
	// we just fiddle around with packets but don't consume them	
	return false;
}

#pragma mark === Outgoing Packet Handling ===
/*----------------------------------------------------------------------------------------
	 OutTranslateIpPacket
------------------------------------------------------------------------------------------*/
// see if our packets our going out of our local network and if so, 
// change the packet's source address and port

void NATHandler::OutTranslateIpPacket(BNetPacket* pkt)
{
	// get the protocol
	uint8 protocol = 0x0;
	pkt->Read(IP_PROTOCOL_OFFSET,  (char*)&protocol, sizeof(protocol) );

	// no need to do any further processing if we don't recognize the protocol
	if( (protocol != IP_PROTO_TCP) && (protocol != IP_PROTO_UDP) ) return;

	// destination IP address
	uint32 dst_addr;
	pkt->Read(IP_DST_IP_OFFSET, (char*)&dst_addr, sizeof(dst_addr) );

	// if destination address is on our subnet, no need 
	// to do anything else, just forward it along.
	if( ((mLocalAddr & mLocalMask) == (dst_addr & mLocalMask))
		|| ntohl(dst_addr) == 0xffffff )  return;
		
	// source IP address
	uint32 src_addr;
	pkt->Read(IP_SRC_IP_OFFSET,  (char*)&src_addr, sizeof(src_addr) );
	
	// just to be safe, make sure the source is from our subnet
	if( (mLocalAddr & mLocalMask) != (src_addr & mLocalMask) ) return;

	// our internet ip address
	uint32 addr = GetInternetAddress();
	if(addr == 0) return;

	// get header_length
	uint8 x = 0;
	pkt->Read(IP_LEN_VERSION_OFFSET,  (char*)&x, sizeof(x) );
	pkt->SetBase(IP_HEADER_LENGTH(x));

	// dispatch to appopriate handler
	switch(protocol)
	{
		case IP_PROTO_TCP:
		case IP_PROTO_UDP:
			OutTranslatePacket(pkt, src_addr, addr, protocol);
		break;
	}

	pkt->SetBase(-IP_HEADER_LENGTH(x));
	
	// replace original source address with our address
	pkt->Write(IP_SRC_IP_OFFSET, (char*)&addr, sizeof(addr) );

	// adjust the checksum
	uint16 checksum = 0;
	pkt->Read(IP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );

	checksum_adjust((uint8*)&checksum, (uint8*)&src_addr, sizeof(src_addr), (uint8*)&addr, sizeof(addr) );

	pkt->Write(IP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );
}

/*----------------------------------------------------------------------------------------
	 OutTranslatePacket
------------------------------------------------------------------------------------------*/
// change the port number of packet
void NATHandler::OutTranslatePacket(BNetPacket* pkt, uint32 orig_src_addr, uint32 new_src_addr, uint8 proto)
{
	// get old port
	uint16 port = 0;
	pkt->Read((proto == IP_PROTO_TCP) ? TCP_SRC_PORT_OFFSET : UDP_SRC_PORT_OFFSET, (char*)&port, sizeof(port) );

	// look up new port
	uint16 new_port = LookupPort(orig_src_addr, port, (proto == IP_PROTO_TCP) ? 0 : 1);
	if(new_port == 0) return;

	// get checksum
	uint16 checksum = 0;
	pkt->Read((proto == IP_PROTO_TCP) ? TCP_CHECKSUM_OFFSET : UDP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );

	// handle funky protocols
	pkt->SetBase((proto == IP_PROTO_TCP) ? TCP_HEADER_SIZE : UDP_HEADER_SIZE );
	for(int i = 0; funky_protocols[i].out_translate_packet; i++)
	{
		if(funky_protocols[i].out_translate_packet(pkt, &checksum, orig_src_addr, new_src_addr, port, new_port))
			break;
	}
	pkt->SetBase(-((proto == IP_PROTO_TCP) ? TCP_HEADER_SIZE : UDP_HEADER_SIZE) );
	
	// replace port number with new port
	pkt->Write((proto == IP_PROTO_TCP) ? TCP_SRC_PORT_OFFSET : UDP_SRC_PORT_OFFSET, (char*)&new_port, sizeof(new_port) );

	// recompute checksum
	checksum_adjust((uint8*)&checksum, (uint8*)&port, sizeof(port), (uint8*)&new_port, sizeof(new_port) );	
	
	// since the checksum is computed with a funky 96-bit pseudo-header that includes the
	// source and destination ip addresses, we've got to do this too
	checksum_adjust((uint8*)&checksum, (uint8*)&orig_src_addr, sizeof(orig_src_addr), (uint8*)&new_src_addr, sizeof(new_src_addr) );

	pkt->Write((proto == IP_PROTO_TCP) ? TCP_CHECKSUM_OFFSET : UDP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );

	if(proto == IP_PROTO_TCP && config.check_fin_flag) CheckFINFlag(pkt, new_port);
}

#pragma mark === Incoming Packet Handling ===

/*----------------------------------------------------------------------------------------
	 InTranslateIpPacket
------------------------------------------------------------------------------------------*/
// look up the real destination address from the port number
void NATHandler::InTranslateIpPacket(BNetPacket* pkt)
{
	// get the protocol
	uint8 protocol = 0x0;
	pkt->Read(IP_PROTOCOL_OFFSET, (char*)&protocol, sizeof(protocol) );
		
	// no need to do any further processing if we don't recognize the protocol
	if( (protocol != IP_PROTO_TCP) && (protocol != IP_PROTO_UDP) ) return;
	
	// destination IP address
	uint32 dst_addr;
	pkt->Read(IP_DST_IP_OFFSET, (char*)&dst_addr, sizeof(dst_addr) );

	// our internet ip address
	uint32 addr = GetInternetAddress();
	if((addr == 0) || (addr != dst_addr)) return;

	// get header_length
	uint8 x = 0;
	pkt->Read(IP_LEN_VERSION_OFFSET,  (char*)&x, sizeof(x) );
	pkt->SetBase(IP_HEADER_LENGTH(x));
	
	switch(protocol)
	{
		case IP_PROTO_TCP:
		case IP_PROTO_UDP:
			addr = InTranslatePacket(pkt, dst_addr, protocol);
		break;
	}
	
	pkt->SetBase(-IP_HEADER_LENGTH(x));

	if(addr == 0) return;	
	
	// replace original destination address with local machine's address
	pkt->Write(IP_DST_IP_OFFSET, (char*)&addr, sizeof(addr) );
		
	// adjust the checksum
	uint16 checksum = 0;
	pkt->Read(IP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );
	
	checksum_adjust((uint8*)&checksum, (uint8*)&dst_addr, sizeof(dst_addr), (uint8*)&addr, sizeof(addr) );
	
	pkt->Write(IP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );
}

/*----------------------------------------------------------------------------------------
	 InTranslatePacket
------------------------------------------------------------------------------------------*/
// returns actual destination ip
uint32 NATHandler::InTranslatePacket(BNetPacket* pkt, uint32 orig_dst_addr, uint8 proto)
{
	// get the port
	uint16 port = 0;
	pkt->Read((proto == IP_PROTO_TCP) ? TCP_DST_PORT_OFFSET : UDP_DST_PORT_OFFSET, (char*)&port, sizeof(port) );
	
	// look up real destination and port
	pair<uint32, uint16> addr_port = LookupAddrPort(port, (proto == IP_PROTO_TCP) ? 0 : 1);

	uint32 new_dst_addr = addr_port.first;
	if(new_dst_addr == 0) return 0;
	uint16 new_port = addr_port.second;
	
	// get checksum
	uint16 checksum = 0;
	pkt->Read((proto == IP_PROTO_TCP) ? TCP_CHECKSUM_OFFSET : UDP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );
	
	// handle funky protocols
	pkt->SetBase((proto == IP_PROTO_TCP) ? TCP_HEADER_SIZE : UDP_HEADER_SIZE);
	for(int i = 0; funky_protocols[i].in_translate_packet; i++)
	{
		if(funky_protocols[i].in_translate_packet(pkt, &checksum, orig_dst_addr, new_dst_addr, port, addr_port.second))
			break;
	}
	pkt->SetBase(-((proto == IP_PROTO_TCP) ? TCP_HEADER_SIZE : UDP_HEADER_SIZE) );
	
	// replace the destination port	
	pkt->Write((proto == IP_PROTO_TCP) ? TCP_DST_PORT_OFFSET : UDP_DST_PORT_OFFSET, (char*)&new_port, sizeof(new_port) );
		
	// recompute checksum
	checksum_adjust((uint8*)&checksum, (uint8*)&port, sizeof(port), (uint8*)&new_port, sizeof(new_port) );	
	
	// since the checksum is computed with a funky 96-bit pseudo-header that includes the
	// source and destination ip addresses, we've got to do this too
	checksum_adjust((uint8*)&checksum, (uint8*)&orig_dst_addr, sizeof(orig_dst_addr), (uint8*)&new_dst_addr, sizeof(new_dst_addr) );

	pkt->Write((proto == IP_PROTO_TCP) ? TCP_CHECKSUM_OFFSET  : UDP_CHECKSUM_OFFSET, (char*)&checksum, sizeof(checksum) );
	
	//if(proto == IP_PROTO_TCP)  CheckFINFlag(pkt, port);

	// return actual destination address
	return new_dst_addr;
}

#pragma mark === Addresses ===

/*----------------------------------------------------------------------------------------
	 FindPPPAddress
------------------------------------------------------------------------------------------*/

void NATHandler::FindPPPAddress(BNetPacket* pkt)
{
	pkt->SetBase(PPP_HEADER_LENGTH);
	
	uint8 code;
	// check for a NAK
	pkt->Read(LCP_CODE_OFFSET, (char*)&code, sizeof(code));
	
	if(code == LCP_CONFIG_NAK)
	{ // and down we go...
		pkt->SetBase(LCP_DATA_OFFSET);
		
		pkt->Read(IPCP_TYPE_OFFSET, (char*)&code, sizeof(code));
	
		if(code == IPCP_IP_ADDR) // and finally the IP address!
			pkt->Read(IPCP_IP_ADDR_OFFSET, (char*)&mInetAddr, sizeof(mInetAddr));
	
		pkt->SetBase(-LCP_DATA_OFFSET);
	}

	pkt->SetBase(-PPP_HEADER_LENGTH);
}

/*----------------------------------------------------------------------------------------
	 GetInternetAddress
------------------------------------------------------------------------------------------*/

uint32 NATHandler::GetInternetAddress()
{
	return mInetAddr;
}

#pragma mark === Address Mapping ===

/*----------------------------------------------------------------------------------------
	 GetFreePort
------------------------------------------------------------------------------------------*/

uint16 NATHandler::GetFreePort(uint8 proto)
{
	int s = -1;
	uint16 port = 0;
	
	{ // get a socket from the pool
		BAutolock lock(mSocketLock[proto]);
		if(mSocketPool[proto].size() > 0)
		{	
			s = mSocketPool[proto].front().first;
			port = mSocketPool[proto].front().second;
			mSocketPool[proto].pop_front();
		}
		
		if(mSocketPool[proto].size() < config.socket_pool_size)
		{
			// wake the great maker up and get some more ports
			resume_thread(mTheGreatMaker);
		}
	}
	
	// get port associated with socket
	
	if(s >= 0)
	{
		BAutolock lock(mPortSocketLock[proto]);
		mPortSocketMap[proto][port] = pair<int, bigtime_t>(s, system_time());
	}
	return port;
}

/*----------------------------------------------------------------------------------------
	 LookupPort
------------------------------------------------------------------------------------------*/
// look up port number associated with IP Address and port
uint16 NATHandler::LookupPort(uint32 addr, uint16 port, uint8 proto)
{	
	BAutolock lock(mPortMapLock[proto]);
	BAutolock lock2(mPortSocketLock[proto]);
	
	pair<uint32, uint16> apair(addr,port);
	
	uint16 new_port = mPortMap[proto][apair];

	// if a port number doesn't exist for this address and port combination, assign one
	if(new_port == 0)
	{ 
		new_port = GetFreePort(proto);
		
		if(new_port > 0)
		{
			mPortMap[proto][apair] = new_port;
			mReversePortMap[proto][new_port] = apair;
		}
	}
	
	if(new_port > 0)
	{
		// update last used time
		pair<int, bigtime_t>& p = mPortSocketMap[proto][new_port];
		if(p.second != STATIC_PORT_MAP)
			p.second = system_time();
	}

	return new_port;
}

/*----------------------------------------------------------------------------------------
	 LookupAddrPort
------------------------------------------------------------------------------------------*/
// look up address and port associated with port
pair<uint32, uint16> NATHandler::LookupAddrPort(uint16 port, uint8 proto)
{	
	BAutolock lock1(mPortMapLock[proto]);
	BAutolock lock2(mPortSocketLock[proto]);
	
	pair<uint32, uint16> p1 = mReversePortMap[proto][port];
	
	// update last used time
	pair<int, bigtime_t>& p2 = mPortSocketMap[proto][port];
	if(p2.second != STATIC_PORT_MAP)
		p2.second = system_time();
	
	return p1;
}

#pragma mark === Threads Galore! ===
/*----------------------------------------------------------------------------------------
	 SocketReaper
------------------------------------------------------------------------------------------*/
// kills old sockets in mSocketPool
int32 NATHandler::SocketReaper(void* data)
{
	NATHandler* h = (NATHandler*)data;
	
	for(;;)
	{	
		if(!done) snooze(config.socket_timeout * MINUTE_MS);	
		if(done) break;
		
		for(int i = 0; i < NAT_NUM_PROTO; i++)
		{
			BAutolock lock(h->mSocketLock[i]);
			while(h->mSocketPool[i].size() > config.socket_pool_size)
			{
				int s = h->mSocketPool[i].front().first;
				h->mSocketPool[i].pop_front();
				if(!config.disable_socket_open)
					closesocket(s);
			}
		}
	}	
	return 0;
}

/*----------------------------------------------------------------------------------------
	 PortReaper
------------------------------------------------------------------------------------------*/
// kills old ports in mPortSocketMap (transfers sockets back to mSocketPool)
int32 NATHandler::PortReaper(void* data)
{	
	NATHandler* h = (NATHandler*)data;
	
	for(;;)
	{	
		if(!done) snooze(config.port_timeout * MINUTE_MS);
		if(done) break;
		
		for(int i = 0; i < NAT_NUM_PROTO; i++)
		{
			BAutolock lock1(h->mPortMapLock[i]);
			BAutolock lock2(h->mPortSocketLock[i]);
			
			// this needs to be more efficient. need to use a list 
			// sorted by last time used or something
			map<uint16, pair<int, bigtime_t> >::iterator it = h->mPortSocketMap[i].begin();
			bigtime_t now = system_time();

			while(it != h->mPortSocketMap[i].end())
			{
				map<uint16, pair<int, bigtime_t> >::iterator cur = it;
				it++;
				bigtime_t last_time = (*cur).second.second;
				if((last_time != STATIC_PORT_MAP) && ((now - last_time) > config.port_timeout * MINUTE_MS)) // socket has expired
				{
					int s = (*cur).second.first;
					uint16 port = (*cur).first;

					// take out of maps
					h->mPortSocketMap[i].erase(cur);
					h->mPortMap[i].erase(h->mReversePortMap[i][port]);
					h->mReversePortMap[i].erase(port);
					
					// put in socket pool
					BAutolock lock3(h->mSocketLock[i]);
					h->mSocketPool[i].push_front(pair<int,uint16>(s, port));
				}
			}
		}
	}	
	
	return 0;
}

/*----------------------------------------------------------------------------------------
	 TheGreatMaker
------------------------------------------------------------------------------------------*/
// gets unused ports for us (puts in mSocketPool)
// we just wake it up everytime we want a new block
// should do this by blocking on a semaphore instead
// of waking up
int32 NATHandler::TheGreatMaker(void* data)
{	
	NATHandler* h = (NATHandler*)data;
	
	// we snooze a little while networking is started
	snooze(5 * 1000000);
	
	h->AddStaticPortMaps();
	
	for(;;)
	{
		// get a block of sockets
		for(int j = 0; j < NAT_NUM_PROTO; j++)
		{
			int s = -1;
			int port = 0;
			struct sockaddr_in sa; 
			memset(sa.sin_zero, 0, sizeof(sa.sin_zero)); 
			{	
				BAutolock lock(h->mSocketLock[j]);	
				for(int i = h->mSocketPool[j].size(); i < config.socket_pool_size; )
				{	
					// HACK
					// open up and bind a socket to safely acquire unused port numbers
					// this really isn't very good, becuase we wind up creating a socket (with its
					// associated thread) for every mapping
					// i'm not sure how safe it is calling socket and bind from within a network add-on,
					// but i haven't had any trouble so far.		
					h->mFreePort[j]++;
					if(h->mFreePort[j] > config.end_port) // wrap around
						h->mFreePort[j] += -config.end_port + config.start_port;
		
					if(config.disable_socket_open)
					{
						i++;
						h->mSocketPool[j].push_back(pair<int, uint16>(0, port));
					}else
					{
						if( (s = socket(AF_INET, (j ? SOCK_STREAM : SOCK_DGRAM), 0)) >= 0)
						{
							sa.sin_family = AF_INET; 				
							port = htons(h->mFreePort[j]);
							sa.sin_port = port; 
							
							sa.sin_addr.s_addr = h->mInetAddr; 
						
							// try to bind the socket
							if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) 
							{
								port = 0;
								closesocket(s);
							}
							else
							{
								i++;
								h->mSocketPool[j].push_back(pair<int, uint16>(s, port));
							}
						}
					}
				}
			}
		}
		// go to sleep
		if(!done) suspend_thread(h->mTheGreatMaker);
		if(done) break;
	}
	
	return 0;
}

#pragma mark === Static Port Mapping ===
/*----------------------------------------------------------------------------------------
	 AddStaticPortMaps
------------------------------------------------------------------------------------------*/
// All we have to do with static port maps in put in a map with an inifinite timeout
void NATHandler::AddStaticPortMaps()
{	
	for(int i = 0; i < NAT_NUM_PROTO; i++)
	{
		mPortMapLock[i].Lock();
		mPortSocketLock[i].Lock();
		mSocketLock[i].Lock();
	}
	
	struct sockaddr_in sa; 
	memset(sa.sin_zero, 0, sizeof(sa.sin_zero)); 

	for(list<port_map>::iterator it = config.static_map.begin(); it != config.static_map.end(); it++)
	{
		uint16 local_port = (*it).local_start_port;
		for(uint16 port = (*it).start_port; port <= (*it).end_port; port++, local_port++)
		{
			for(int proto = 0; proto < NAT_NUM_PROTO; proto++)
			{
				if(!((*it).port_type & (1 << proto))) continue;
								
				uint16 nport = htons(port);

				if(config.disable_socket_open)
				{
					mPortSocketMap[proto][nport] = pair<int, bigtime_t>(0,STATIC_PORT_MAP);
					pair<uint32, uint16> ip_port((*it).ip_addr, htons(local_port));
					mPortMap[proto][ip_port] = nport;
					mReversePortMap[proto][nport] = ip_port;
				}else
				{ // bind a socket on this port so nobody else can use it

					int s;
					if( (s = socket(AF_INET, (proto ? SOCK_STREAM : SOCK_DGRAM), 0)) >= 0)
					{
						sa.sin_family = AF_INET; 				
						sa.sin_port = nport; 
						
						sa.sin_addr.s_addr = mInetAddr; 
					
						if(bind(s, (struct sockaddr *)&sa, sizeof(sa)) >= 0)
						{
							mPortSocketMap[proto][nport] = pair<int, bigtime_t>(s,STATIC_PORT_MAP);
							pair<uint32, uint16> ip_port((*it).ip_addr, htons(local_port));
							mPortMap[proto][ip_port] = nport;
							mReversePortMap[proto][nport] = ip_port;
						}else
							closesocket(s);
					}
				}
			}
		}
	}
	
	for(int i = 0; i < NAT_NUM_PROTO; i++)
	{
		mPortMapLock[i].Unlock();
		mPortSocketLock[i].Unlock();
		mSocketLock[i].Unlock();
	}
}

#pragma mark === ===
/*----------------------------------------------------------------------------------------
	 CheckFINFlag
------------------------------------------------------------------------------------------*/
// check see see if the connection is closing. if so, we can reclaim the port
void NATHandler::CheckFINFlag(BNetPacket* pkt,uint16 port)
{	
	uint8 flags;
	pkt->Read(TCP_FLAGS_OFFSET, (char*)&flags, sizeof(flags) );
	if(flags & FIN_MASK) 
	{ // take out of port map and put back in socket pool

		BAutolock lock1(mPortMapLock[0]);
		BAutolock lock2(mPortSocketLock[0]);

		pair<int, bigtime_t> p = mPortSocketMap[0][port];
		if(p.second != STATIC_PORT_MAP)
		{
			int s = p.first;
			
			// take out of maps
			mPortSocketMap[0].erase(port);
			mPortMap[0].erase(mReversePortMap[0][port]);
			mReversePortMap[0].erase(port);

			// put in socket pool
			mSocketPool[0].push_front(pair<int,uint16>(s, port));
		}
	}
}

#pragma mark -
#pragma mark === checksum manipulation ===
/*----------------------------------------------------------------------------------------
	 checksum_adjust
------------------------------------------------------------------------------------------*/
// adapted from RFC 1631, which didn't work worth a shit originally
void checksum_adjust(uint8* chksum, uint8 *optr, int32 olen, uint8 *nptr, int32 nlen)
/* assuming: unsigned char is 8 bits, long is 32 bits.
  - chksum points to the chksum in the packet
  - optr points to the old_val data in the packet
  - nptr points to the new data in the packet
*/
{
	uint16 x, last, old_val, new_val;
	x = chksum[0] * 0x100 + chksum[1];
	x = ~x;
	
	last = x;
	while (olen) 
	{
		if (olen==1) 
		{
			old_val = optr[0] * 0x100 + optr[1];
			x -= old_val & 0xff00;
			if (x > last) x--; 
			break;
		}else 
		{
			old_val = optr[0] * 0x100 + optr[1]; 
			optr += 2;
			x -= old_val;
			if (x > last) x--; 
			olen -= 2;
		}
		last = x;
	}

	last = x;
	while (nlen) 
	{
		if (nlen == 1) 
		{
			new_val = nptr[0] * 0x100 + nptr[1];
			x += new_val & 0xff00;
			if (x < last) x++;
			break;
		}
		else {
			new_val = nptr[0] * 0x100 + nptr[1]; 
			nptr+=2;
			x += new_val;
			if (x < last) x++;
			nlen -= 2;
		}
		last = x;
	}

	x = ~x;
	chksum[0]= ((x >> 8) & 0xff); chksum[1] = (x & 0xff);
}

#pragma mark -
#pragma mark === export ===
/*----------------------------------------------------------------------------------------
	 open_protocol
------------------------------------------------------------------------------------------*/

BNetProtocol * open_protocol(const char *device)
{	
	memset(&STATIC_PORT_MAP,0xff,sizeof(STATIC_PORT_MAP));

	// debugging
//	f = fopen("/tmp/net.dump", "w+");

	config.load_settings();
	
	return new NATHandler();
}
